# MIT License

# Copyright (c) 2016 Diogo Dutra

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from falconswagger.models.base import get_model_schema
from falconswagger.exceptions import ModelBaseError
from myreco.engines.types.base import EngineTypeChooser
from myreco.engines.types.items_indices_map import ItemsIndicesMap
from myreco.items_types.models import build_item_key
from types import MethodType, FunctionType
from jsonschema import ValidationError
from sqlalchemy.ext.declarative import AbstractConcreteBase, declared_attr
from importlib import import_module
from falcon import HTTPNotFound
import sqlalchemy as sa
import json


class EnginesModelBase(AbstractConcreteBase):
    __tablename__ = 'engines'
    __schema__ = get_model_schema(__file__)
    _jobs = dict()

    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    name = sa.Column(sa.String(255), unique=True, nullable=False)
    configuration_json = sa.Column(sa.Text, nullable=False)

    @property
    def configuration(self):
        if not hasattr(self, '_configuration'):
            self._configuration = json.loads(self.configuration_json)
        return self._configuration

    @declared_attr
    def store_id(cls):
        return sa.Column(sa.ForeignKey('stores.id'), nullable=False)

    @declared_attr
    def type_name_id(cls):
        return sa.Column(sa.ForeignKey('engines_types_names.id'), nullable=False)

    @declared_attr
    def item_type_id(cls):
        return sa.Column(sa.ForeignKey('items_types.id'), nullable=False)

    @declared_attr
    def type_name(cls):
        return sa.orm.relationship('EnginesTypesNamesModel')

    @declared_attr
    def item_type(cls):
        return sa.orm.relationship('ItemsTypesModel')

    @declared_attr
    def store(cls):
        return sa.orm.relationship('StoresModel')

    @property
    def type_(self):
        if not hasattr(self, '_type'):
            self._set_type()
        return self._type

    def _set_type(self):
        todict_schema = {'variables': False}
        self._type = EngineTypeChooser(self.type_name.name)(self.todict(todict_schema))

    def _setattr(self, attr_name, value, session, input_):
        if attr_name == 'configuration':
            value = json.dumps(value)
            attr_name = 'configuration_json'

        if attr_name == 'type_name_id':
            value = {'id': value}
            attr_name = 'type_name'

        if attr_name == 'item_type_id':
            value = {'id': value}
            attr_name = 'item_type'

        super()._setattr(attr_name, value, session, input_)

    def _format_output_json(self, dict_inst, schema):
        if schema.get('configuration') is not False:
            dict_inst.pop('configuration_json')
            dict_inst['configuration'] = self.configuration

        if schema.get('variables') is not False:
            dict_inst['variables'] = self.type_.get_variables()


class EnginesModelDataImporterBase(EnginesModelBase):
    __schema__ = get_model_schema(__file__, 'data_importer_schema.json')

    @classmethod
    def _run_job(cls, job_session, req, resp):
        engine = cls._get_engine(req, job_session)
        try:
            data_importer = import_module(engine.configuration['data_importer_path'])
        except Exception as error:
            raise ModelBaseError("invalid 'data_importer_path' configuration for this engine")

        items_indices_map = cls._build_items_indices_map(job_session, engine)
        return data_importer.get_data(engine, items_indices_map)

    @classmethod
    def _get_engine(cls, req, job_session):
        id_ = req.context['parameters']['uri_template']['id']
        engines = cls.get(job_session, {'id': id_}, todict=False)

        if not engines:
            raise HTTPNotFound()

        return engines[0]

    @classmethod
    def _build_items_indices_map(cls, job_session, engine):
        item_collection_model = cls.__api__.models[build_item_key(engine.item_type.name)]
        item_model = item_collection_model.__models__[engine.store_id]
        return ItemsIndicesMap(job_session, item_model)


class EnginesModelObjectsExporterBase(EnginesModelDataImporterBase):
    __schema__ = get_model_schema(__file__, 'objects_exporter_schema.json')

    @classmethod
    def _run_job(cls, job_session, req, resp):
        import_data = req.context['parameters']['query_string'].get('import_data')
        engine = cls._get_engine(req, job_session)
        items_indices_map = cls._build_items_indices_map(job_session, engine)

        if import_data:
            try:
                data_importer = import_module(engine.configuration['data_importer_path'])
            except ImportError:
                raise ModelBaseError("Invalid 'data_importer_path'", input_=engine.configuration)

            data_importer.get_data(engine, items_indices_map)
            return engine.type_.export_objects(job_session, items_indices_map)
        else:
            return engine.type_.export_objects(job_session, items_indices_map)


class EnginesTypesNamesModelBase(AbstractConcreteBase):
    __tablename__ = 'engines_types_names'
    __use_redis__ = False

    id = sa.Column(sa.Integer, primary_key=True)
    name = sa.Column(sa.String(255), unique=True, nullable=False)



# MIT License

# Copyright (c) 2016 Diogo Dutra

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class ItemsIndicesDict(dict):

    def __init__(self, items_indices_map, items_model):
        self._items_model = items_model
        dict.__init__(self, items_indices_map)

    def get(self, keys, default=None):
        key = self._items_model(keys).get_key().encode()
        value = dict.get(self, key, default)
        if value is not default:
            return int(value.decode())


class ItemsIndicesMap(object):

    def __init__(self, session, items_model):
        self.session = session
        self.items_model = items_model

    def get_all(self):
        items_indices_map = self.session.redis_bind.hgetall(self._build_key())
        return ItemsIndicesDict(items_indices_map, self.items_model)

    def _build_key(self):
        return self.items_model.__key__ + '_indices_map'

    def get_indices_items_map(self):
        map_ = self.session.redis_bind.hgetall(self._build_indices_items_key())
        return {int(k): v.decode() for k, v in map_.items()}

    def _build_indices_items_key(self):
        return self.items_model.__key__ + '_items_map'

    def update(self):
        redis_key = self._build_key()
        indices_items_key = self._build_indices_items_key()
        items_indices_map = self.session.redis_bind.hgetall(redis_key)
        indices_items_map = self.session.redis_bind.hgetall(indices_items_key)

        items = self.items_model.get(self.session)
        items_keys = set([self.items_model(item).get_key().encode() for item in items])

        new_keys = [key for key in items_keys if key not in items_indices_map]
        old_keys = set([key for key in items_keys if key in items_indices_map])
        keys_to_delete = set(items_indices_map.keys()).difference(old_keys)
        free_indices = [int(v) for k, v in items_indices_map.items() if k in keys_to_delete]

        [items_indices_map.pop(k) for k in keys_to_delete]
        [indices_items_map.pop(i) for i in free_indices]

        if old_keys:
            iterable = (int(v) for k, v in items_indices_map.items() if k in old_keys)
            counter = max(iterable)
        else:
            counter = 0

        i = 0
        for i, key, index in enumerate(zip(new_keys, free_indices)):
            items_indices_map[key] = index
            indices_items_map[index] = key

        for key in new_keys[i:]:
            items_indices_map[key] = counter
            indices_items_map[counter] = key
            counter += 1

        if keys_to_delete:
            self.session.redis_bind.hdel(redis_key, *keys_to_delete)
            self.session.redis_bind.hdel(indices_items_key, *free_indices)

        if items_indices_map:
            self.session.redis_bind.hmset(redis_key, items_indices_map)
            self.session.redis_bind.hmset(indices_items_key, indices_items_map)

        return self._format_output(self.get_all())

    def _format_output(self, output):
        return {' | '.join([str(i) for i in eval(k)]): int(v.decode()) for k, v in output.items()}

    def get_items(self, session, indices):
        return [item.decode() for item in \
            session.redis_bind.hmget(self._build_indices_items_key(), indices) if item is not None]



# MIT License

# Copyright (c) 2016 Diogo Dutra

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from falconswagger.models.base import build_validator, get_module_path
from jsonschema import Draft4Validator
from abc import ABCMeta, abstractmethod
from bottleneck import argpartition
import msgpack


class EngineError(Exception):
    pass


class EngineTypeMeta(type):

    def __init__(cls, name, bases_classes, attributes):
        if name != 'EngineType':
            schema = cls.__configuration_schema__
            Draft4Validator.check_schema(schema)
            cls.__config_validator__ = build_validator(schema, get_module_path(cls))


class EngineType(metaclass=EngineTypeMeta):

    def __init__(self, engine=None, items_model=None):
        if engine is not None:
            self.validate_config(engine)

        self.engine = engine
        self.items_model = items_model

    def get_variables(self):
        return []

    def validate_config(self, engine):
        self.__config_validator__.validate(engine['configuration'])
        self._validate_config(engine)

    def _validate_config(self, engine):
        pass

    def get_recommendations(self, session, filters, max_recos, **variables):
        rec_vector = self._build_rec_vector(session, **variables)
        [filter_.filter(session, rec_vector, ids) for filter_, ids in filters.items()]
        return self._build_rec_list(session, rec_vector, max_recos)

    def _build_rec_vector(self, session, **variables):
        pass

    def _build_rec_list(self, session, rec_vector, max_recos):
        items_indices_map = self.items_model.build_items_indices_map(session)
        best_indices = self._get_best_indices(rec_vector, max_recos)
        best_items_keys = items_indices_map.get_items(session, best_indices)
        return [msgpack.loads(item, encoding='utf-8') for item in session.redis_bind.hmget(
                            self.items_model.__key__, best_items_keys) if item is not None]

    def _get_best_indices(self, rec_vector, max_recos):
        if max_recos > rec_vector.size:
            max_recos = rec_vector.size

        best_indices = argpartition(rec_vector, max_recos-1)[:max_recos]
        best_values = rec_vector[best_indices]
        return [i for i, _ in
            sorted(zip(best_indices, best_values), key=lambda x: x[1], reverse=True)]

    def export_objects(self, session):
        pass


class AbstractDataImporter(metaclass=ABCMeta):

    @classmethod
    @abstractmethod
    def get_data(cls, engine):
        pass


from myreco.engines.types.neighborhood.engine import NeighborhoodEngine
from myreco.engines.types.top_seller.engine import TopSellerEngine
from myreco.engines.types.visual_similarity.engine import VisualSimilarityEngine


class EngineTypeChooser(object):

    def __new__(cls, name):
        if name == 'neighborhood':
            return NeighborhoodEngine

        elif name == 'top_seller':
            return TopSellerEngine

        elif name == 'visual_similarity':
            return VisualSimilarityEngine



# MIT License

# Copyright (c) 2016 Diogo Dutra

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from glob import glob
from collections import namedtuple
from itertools import starmap
import os.path
import csv
import gzip


def build_csv_readers(path):
    readers = []
    for filename in glob(os.path.join(path, '*.gz')):
        with gzip.open(filename) as csv_file:
            readers.append(csv.DictReader(csv_file))
    return readers


def build_data_path(engine):
    return os.path.join(engine['store']['configuration']['data_path'],
                        engine['type_name']['name'])



# MIT License

# Copyright (c) 2016 Diogo Dutra

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.engines.types.base import EngineType, EngineError
from myreco.engines.types.utils import build_csv_readers, build_data_path
from falconswagger.models.base import get_model_schema
from concurrent.futures import ThreadPoolExecutor
import numpy as np
import os.path
import zlib


class TopSellerEngine(EngineType):
    __configuration_schema__ = get_model_schema(__file__)

    def export_objects(self, session, items_indices_map):
        readers = build_csv_readers(build_data_path(self.engine))
        top_seller_vector = self._build_top_seller_vector(readers, items_indices_map)
        redis_key = self._build_redis_key()
        session.redis_bind.set(redis_key, zlib.compress(top_seller_vector.tobytes()))
        result = sorted(enumerate(top_seller_vector), key=(lambda x: (x[1], x[0])), reverse=True)
        indices_items_map = items_indices_map.get_indices_items_map()
        return [{self._format_output(indices_items_map, r): int(r[1])} for r in result]

    def _format_output(self, indices_items_map, r):
        return ' | '.join(eval(indices_items_map[r[0]]))

    def _build_top_seller_vector(self, readers, items_indices_map):
        error_message = "No data found for engine '{}'".format(self.engine['name'])
        if not len(readers):
            raise EngineError(error_message)

        executor = ThreadPoolExecutor(len(readers))
        jobs = []
        indices_values_map = dict()
        for reader in readers:
            job = executor.submit(self._set_indices_values_map,
                                indices_values_map, reader, items_indices_map)
            jobs.append(job)

        [job.result() for job in jobs]

        if not indices_values_map:
            raise EngineError(error_message)

        vector = np.zeros(max(indices_values_map.keys())+1, dtype=np.int32)
        indices = np.array(list(indices_values_map.keys()), dtype=np.int32)
        vector[indices] = np.array(list(indices_values_map.values()), dtype=np.int32)

        return vector

    def _set_indices_values_map(self, indices_values_map, reader, items_indices_map):
        items_indices_map = items_indices_map.get_all()

        for line in reader:
            value = line.pop('value')
            index = items_indices_map.get(line)
            if index is not None:
                indices_values_map[int(index)] = int(value)

    def _build_redis_key(self):
        return '{}_{}'.format(self.engine['type_name']['name'], self.engine['id'])



# MIT License

# Copyright (c) 2016 Diogo Dutra

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from myreco.engines.types.items_indices_map import ItemsIndicesMap
from falconswagger.models.redis import RedisModelMeta, RedisModelBuilder
from falconswagger.models.base import get_model_schema, ModelBase, get_dir_path
from falconswagger.exceptions import ModelBaseError
from sqlalchemy.ext.declarative import AbstractConcreteBase, declared_attr
from jsonschema import ValidationError, Draft4Validator
from copy import deepcopy
from falcon import HTTPNotFound
import sqlalchemy as sa
import json


def build_item_key(name, store_id=None):
    name = name.lower().replace(' ', '_')
    if store_id:
        return '{}_{}'.format(name, store_id)
    return name


class ItemsTypesModelBase(AbstractConcreteBase):
    __tablename__ = 'items_types'
    __schema__ = get_model_schema(__file__)
    __build_items_models__ = True

    id = sa.Column(sa.Integer, primary_key=True)
    name = sa.Column(sa.String(255), unique=True, nullable=False)
    schema_json = sa.Column(sa.Text, nullable=False)

    @declared_attr
    def stores(cls):
        return sa.orm.relationship('StoresModel', uselist=True, secondary='items_types_stores')

    def _setattr(self, attr_name, value, session, input_):
        if attr_name == 'schema':
            self._validate_input(value)
            value = json.dumps(value)
            attr_name = 'schema_json'

        super()._setattr(attr_name, value, session, input_)

    def _validate_input(self, schema):
        Draft4Validator.check_schema(schema)

        for id_name in schema['id_names']:
            if id_name not in schema.get('properties', {}):
                raise ValidationError(
                    "id_name '{}' was not found in schema properties".format(id_name),
                    instance=schema['id_names'], schema=schema)

    def _format_output_json(self, dict_inst, todict_schema):
        if todict_schema.get('schema') is not False:
            if 'schema_json' in dict_inst:
                dict_inst['schema'] = json.loads(dict_inst.pop('schema_json'))

                schema_properties = dict_inst['schema'].get('properties', {})
                schema_properties_names = sorted(schema_properties.keys())
                dict_inst['available_filters'] = \
                    [{'name': name, 'schema': schema_properties[name]} \
                        for name in schema_properties_names]

    @classmethod
    def associate_all_items(cls, session):
        items_types = cls.get(session)
        cls.associate_items(items_types)

    @classmethod
    def associate_items(cls, items_types):
        [cls._build_items_collections_model(item_type) for item_type in items_types]

    @classmethod
    def _build_items_collections_model(cls, item_type):
        if cls.__api__:
            name, schema = item_type['name'], item_type['schema']
            id_names = schema['id_names']
            class_name = cls._build_class_name(name)
            key = build_item_key(name)
            items_models = {store['id']: cls._build_item_model(item_type, store) \
                for store in item_type['stores']}
            attributes = {
                '__key__': key,
                '__schema__': cls._build_items_collections_schema(key, schema, id_names),
                '__models__': items_models
            }
            items_collections = ItemsCollectionsModelBaseMeta(class_name, (ModelBase,), attributes)
            cls.__api__.associate_model(items_collections)

    @classmethod
    def _build_class_name(cls, *names):
        final_name = ''
        for name in names:
            name = name.split(' ')
            for in_name in name:
                final_name += in_name.capitalize()

        return final_name + 'Model'

    @classmethod
    def _build_item_model(cls, item_type, store):
        class_name = cls._build_class_name(item_type['name'], store['name'])
        key = build_item_key(item_type['name'], store['id'])
        id_names = item_type['schema']['id_names']
        return RedisModelBuilder(class_name, key, id_names, None, metaclass=ItemsModelBaseMeta)

    @classmethod
    def _build_items_collections_schema(cls, key, schema, id_names):
        base_uri = '/{}'.format(key)
        id_names_uri = base_uri + '/' + '/'.join(['{{{}}}'.format(id_name) for id_name in id_names])
        patch_schema = deepcopy(schema)
        required = patch_schema.get('required')
        if required:
            patch_schema['required'] = [req for req in required if req in id_names]
        properties = patch_schema.get('properties')
        if properties:
            properties['_operation'] = {'enum': ['delete', 'update']}

        swagger_schema = {
            base_uri: {
                'parameters': [{
                    'name': 'Authorization',
                    'in': 'header',
                    'required': True,
                    'type': 'string'
                },{
                    'name': 'store_id',
                    'in': 'query',
                    'required': True,
                    'type': 'integer'
                }],
                'post': {
                    'parameters': [{
                        'name': 'body',
                        'in': 'body',
                        'required': True,
                        'schema': {
                            'type': 'array',
                            'minItems': 1,
                            'items': schema
                        }
                    }],
                    'operationId': 'post_by_body',
                    'responses': {'201': {'description': 'Created'}}
                },
                'patch': {
                    'parameters': [{
                        'name': 'body',
                        'in': 'body',
                        'required': True,
                        'schema': {
                            'type': 'array',
                            'minItems': 1,
                            'items': patch_schema
                        }
                    }],
                    'operationId': 'patch_by_body',
                    'responses': {'200': {'description': 'Updated'}}
                },
                'get': {
                    'parameters': [{
                        'name': 'page',
                        'in': 'query',
                        'type': 'integer'
                    },{
                        'name': 'items_per_page',
                        'in': 'query',
                        'type': 'integer'
                    }],
                    'operationId': 'get_by_body',
                    'responses': {'200': {'description': 'Got'}}
                },
            },
            id_names_uri: {
                'parameters': [{
                    'name': 'Authorization',
                    'in': 'header',
                    'required': True,
                    'type': 'string'
                },{
                    'name': 'store_id',
                    'in': 'query',
                    'required': True,
                    'type': 'integer'
                }],
                'get': {
                    'operationId': 'get_by_uri_template',
                    'responses': {'200': {'description': 'Got'}}
                }
            }
        }

        base_uri_get_parameters = cls._build_id_names_parameters_schema(id_names, schema, 'query')
        swagger_schema[base_uri]['get']['parameters'].extend(base_uri_get_parameters)

        id_names_uri_parameters = cls._build_id_names_parameters_schema(id_names, schema, 'path')
        swagger_schema[id_names_uri]['parameters'].extend(id_names_uri_parameters)

        return swagger_schema

    @classmethod
    def _build_id_names_parameters_schema(cls, id_names, schema, in_):
        parameters = []
        for id_name in id_names:
            parameter = deepcopy(schema['properties'][id_name])
            parameter.update({
                'name': id_name,
                'in': in_
            })
            if in_ == 'path':
                parameter['required'] = True
            parameters.append(parameter)

        return parameters

    @classmethod
    def insert(cls, session, objs, commit=True, todict=True, **kwargs):
        objs = type(cls).insert(cls, session, objs, commit=commit, todict=todict, **kwargs)
        if cls.__build_items_models__:
            cls.associate_items(objs)

        return objs

    @classmethod
    def update(cls, session, objs, commit=True, todict=True, ids=None, **kwargs):
        old_items_types = cls.get(session, ids=ids)
        objs = type(cls).update(cls, session, objs, commit=commit, todict=todict, ids=ids, **kwargs)
        cls.reassociate_items(old_items_types, objs)
        return objs

    @classmethod
    def reassociate_items(cls, old_items_types, new_items_types):
        new_names = [item_type['name'] for item_type in new_items_types]
        old_items_types = \
            [item_type for item_type in old_items_types if item_type['name'] not in new_names]

        for item_type in old_items_types:
            cls._disassociate_item(item_type)

        for item_type in new_items_types:
            cls._disassociate_item(item_type)
            cls._build_items_collections_model(item_type)

    @classmethod
    def _disassociate_item(cls, item_type):
        name = item_type['name']
        if name in cls.__api__.models:
            cls.__api__.disassociate_model(cls.__api__.models[name])

    @classmethod
    def delete(cls, session, ids, commit=True, **kwargs):
        items_types = cls.get(session, ids=ids)
        type(cls).delete(cls, session, ids, commit=commit, **kwargs)
        [cls._disassociate_item(item_type) for item_type in items_types]

    @staticmethod
    def get_module_path():
        return get_dir_path(__file__)


class ItemsTypesModelIndicesUpdaterBase(ItemsTypesModelBase):

    @classmethod
    def _build_items_collections_schema(cls, key, schema, id_names):
        update_indices_uri = '/{}/update_indices'.format(key)
        schema = ItemsTypesModelBase._build_items_collections_schema(key, schema, id_names)
        schema[update_indices_uri] = {
            'parameters': [{
                'name': 'Authorization',
                'in': 'header',
                'required': True,
                'type': 'string'
            },{
                'name': 'store_id',
                'in': 'query',
                'required': True,
                'type': 'integer'
            }],
            'post': {
                'operationId': 'post_job',
                'responses': {'201': {'description': 'Executing'}}
            },
            'get': {
                'parameters': [{
                    'name': 'hash',
                    'in': 'query',
                    'required': True,
                    'type': 'string'
                }],
                'operationId': 'get_job',
                'responses': {'200': {'description': 'Got'}}
            }
        }
        return schema


class ItemsModelBaseMeta(RedisModelMeta):

    def get(cls, session, ids=None, limit=None, offset=None, **kwargs):
        items_per_page, page = kwargs.get('items_per_page', 1000), kwargs.get('page', 1)
        limit = items_per_page * page
        offset = items_per_page * (page-1)
        return RedisModelMeta.get(cls, session, ids=ids, limit=limit, offset=offset, **kwargs)

    def build_items_indices_map(cls, session):
        return ItemsIndicesMap(session, cls)


class ItemsCollectionsModelBaseMeta(RedisModelMeta):

    def insert(cls, session, objs, **kwargs):
        item_model = cls._get_model(kwargs)
        return item_model.insert(session, objs, **kwargs)

    def _get_model(cls, kwargs):
        store_id = kwargs.pop('store_id')
        item_model = cls.__models__.get(store_id)
        if item_model is None:
            raise HTTPNotFound()
        return item_model

    def update(cls, session, objs, ids=None, **kwargs):
        item_model = cls._get_model(kwargs)
        return item_model.update(session, objs, ids=ids, **kwargs)

    def get(cls, session, ids=None, limit=None, offset=None, **kwargs):
        item_model = cls._get_model(kwargs)
        return item_model.get(session, ids=ids, limit=limit, offset=offset, **kwargs)

    def _run_job(cls, job_session, req, resp):
        params = req.context['parameters']['query_string']
        items_indices_map = ItemsIndicesMap(job_session, cls._get_model(params))
        return items_indices_map.update()


def build_items_types_stores_table(metadata, **kwargs):
    return sa.Table(
        'items_types_stores', metadata,
        sa.Column('item_type_id', sa.Integer, sa.ForeignKey('items_types.id', ondelete='CASCADE', onupdate='CASCADE'), primary_key=True),
        sa.Column('store_id', sa.Integer, sa.ForeignKey('stores.id', ondelete='CASCADE', onupdate='CASCADE'), primary_key=True),
        **kwargs)



# MIT License

# Copyright (c) 2016 Diogo Dutra

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from zlib import decompress
import numpy as np


class FilterBaseBy(object):

    def __init__(self, items_model, name, is_inclusive, id_name=None):
        self.key = items_model.__key__ + '_' + name + '_filter'
        self.items_model = items_model
        self.name = name
        self.is_inclusive = is_inclusive
        self.id_name = id_name

    def filter(self, session, rec_vector):
        filter_ = session.redis_bind.get(self.key)
        if filter_ is not None:
            filter_ = self._prepare_filter(filter_, rec_vector.size)
            self._filter(filter_, rec_vector)

    def _prepare_filter(self, filter_, new_size):
        return self._resize_vector(np.fromstring(decompress(filter_, new_size)))

    def _resize_vector(self, vector, new_size):
        if vector.size != new_size:
            vector.resize(new_size, refcheck=False)

    def _filter(self, filter_, rec_vector):
        if not self.is_inclusive:
            filter_ = np.invert(filter_)

        if np.sum(filter_):
            rec_vector *= filter_


class BooleanFilterBy(FilterBaseBy):
    pass


class MultipleFilterBy(FilterBaseBy):

    def filter(self, session, rec_vector, ids):
        filters = session.redis_bind.hmget(self.key, *ids)
        filters = [self._prepare_filter(filter_. rec_vector.size)
                    for filter_ in filters if filter_ is not None]
        final_filter = np.zeros(rec_vector.size, dtype=np.bool)

        for filter_ in filters:
            final_filter = np.logical_or(final_filter, filter_)

        self._filter(final_filter, rec_vector)


class SimpleFilterBy(MultipleFilterBy):
    pass


class ObjectFilterBy(MultipleFilterBy):

    def _get_id_from_property(self, item):
        return item[self.name][self.id_name]


class ArrayFilterBy(MultipleFilterBy):
    pass


class SimpleFilterOf(SimpleFilterBy):

    def filter(self, session, rec_vector, items_ids):
        items = self.items_model.get(session, items_ids)
        filter_ids = [item[self.name] for item in items]
        SimpleFilterBy.filter(self, session, rec_vector, filter_ids)


class ObjectFilterOf(ObjectFilterBy):

    def filter(self, session, rec_vector, items_ids):
        items = self.items_model.get(session, items_ids)
        filter_ids = [self._get_id_from_property(item) for item in items]
        ObjectFilterBy.filter(self, session, rec_vector, filter_ids)


class ArrayFilterOf(ArrayFilterBy):

    def filter(self, session, rec_vector, items_ids):
        items = self.items_model.get(session, items_ids)
        filter_ids = []
        [filter_ids.extend(item[self.name]) for item in items]
        ArrayFilterBy.filter(self, session, rec_vector, filter_ids)
